

import unittest
import logging

from pimpmyclass import helpers

class TestDoc(unittest.TestCase):

    def assertDocEqual(self, doc1, doc2):
        if doc1 is None or doc2 is None:
            self.assertEqual(doc1, doc2)

        doc1 = '\n'.join(d.strip() for d in doc1.split('\n') if d.strip())
        doc2 = '\n'.join(d.strip() for d in doc2.split('\n') if d.strip())
        self.assertEqual(doc1, doc2)

    def test_indent_spaces(self):
        s4 = "    hello\n    word\n \n    !"
        self.assertEqual(helpers.guess_indent(s4), 4 * ' ')

        s3 = "    hello\n   word\n \n    !"
        self.assertEqual(helpers.guess_indent(s3), 3 * ' ')

    def test_indent_tabs(self):
        s3 = "\t\t\thello\n\t\t\tword\n \n\t\t\t   !"
        self.assertEqual(helpers.guess_indent(s3), 3 * '\t')

        s2 = "\t\thello\n\t\tword\n\n"
        self.assertEqual(helpers.guess_indent(s2), 2 * '\t')

    def test_indent_mixed(self):

        with self.assertRaises(ValueError):
            s3 = "\t\t\thello\n word\n \n\t\t\t   !"
            out = helpers.guess_indent(s3)

    def test_append_prepend(self):

        def f():
            """This is a function

            And here is the rest
            """

        def fp():
            """Async This is a function

            And here is the rest
            """

        def fa():
            """This is a function

            And here is the rest

            a = 1
            b = 2
            """

        self.assertDocEqual(helpers.prepend_to_docstring('Async ', f.__doc__), fp.__doc__)
        self.assertDocEqual(helpers.append_lines_to_docstring(['', 'a = 1', 'b = 2'], f.__doc__), fa.__doc__)





# -*- coding: utf-8 -*-

import unittest
import logging

from pimpmyclass import mixins, dictprops, props, helpers


class MemHandler(logging.Handler):

    def __init__(self):
        super().__init__()
        self.setFormatter(logging.Formatter(style='{'))
        self.history = list()

    def emit(self, record):
        self.history.append(self.format(record))


def define(proptype, *bases, **kwargs):

    class Dummy(*bases, **kwargs):
        _internal = {}

        @proptype
        def test(self, key):
            return self._internal.get(key, None)

        @test.setter
        def test(self, key, value):
            self._internal[key] = value

    return Dummy


def define_w_keys(proptype, keys, *bases, **kwargs):

    class Dummy(*bases, **kwargs):
        _internal = {}

        @proptype(keys=keys)
        def test(self, key):
            return self._internal.get(key, None)

        @test.setter
        def test(self, key, value):
            self._internal[key] = value

    return Dummy


class TestMixins(unittest.TestCase):

    def test_dict(self):

        """
        with self.assertRaises(Exception):
            define(dictprops.DictProperty)
        """

        dummy = define(dictprops.DictProperty)()
        self.assertIsInstance(dummy.test, dictprops.BoundedDictProperty)
        self.assertIs(dummy.test[0], None)
        dummy.test[0] = 4
        self.assertEqual(dummy.test[0], 4)

    def test_dict_cache(self):

        class MyProp(props.PreventUnnecessarySetProperty, props.GetSetCacheProperty):
            pass


        class MyDictProperty(dictprops.DictCacheProperty):

            _subproperty_init = MyProp

        with self.assertRaises(Exception):
            define(dictprops.DictCacheProperty)

        Dummy = define(MyDictProperty, mixins.CacheMixin, mixins.StorageMixin, mixins.BaseLogMixin)

        dummy = Dummy()

        out = dummy.recall('test')
        self.assertEqual(out, {})
        self.assertIsInstance(out, helpers.missingdict)
        # Any non present value
        self.assertEqual(out[1234], Dummy._cache_unset_value)

        self.assertIs(dummy.test[0], None)
        dummy.test[0] = 4

        out = dummy.recall('test')
        self.assertEqual(out, {0: 4})
        self.assertEqual(out[0], 4)
        self.assertIsInstance(out, helpers.missingdict)
        # Any non present value
        self.assertEqual(out[1234], Dummy._cache_unset_value)

        # The storage format is using a (named)tuple
        self.assertEqual(dummy.storage['cache'], {('test', 0): 4})

    def test_dict_dictkeys(self):

        dummy = define_w_keys(dictprops.DictProperty, {'x': 1, 2: 'y'})()
        self.assertIsInstance(dummy.test, dictprops.BoundedDictProperty)

        with self.assertRaises(KeyError):
            dummy.test[1] = 1

        with self.assertRaises(KeyError):
            dummy.test[1]

        with self.assertRaises(KeyError):
            dummy.test['y'] = 1

        with self.assertRaises(KeyError):
            dummy.test['y']

        dummy.test['x'] = 4
        self.assertEqual(dummy.test['x'], 4)
        self.assertEqual(dummy._internal[1], 4)

    def test_dict_enumkeys(self):

        import enum

        class KK(enum.Enum):
            X = 1
            Two = 'y'

        dummy = define_w_keys(dictprops.DictProperty, KK)()
        self.assertIsInstance(dummy.test, dictprops.BoundedDictProperty)

        with self.assertRaises(KeyError):
            dummy.test[1] = 1

        with self.assertRaises(KeyError):
            dummy.test[1]

        with self.assertRaises(KeyError):
            dummy.test['y'] = 1

        with self.assertRaises(KeyError):
            dummy.test['y']

        dummy.test[KK.X] = 4
        self.assertEqual(dummy.test['X'], 4)
        self.assertEqual(dummy._internal[1], 4)

        dummy.test['Two'] = 2
        self.assertEqual(dummy.test['Two'], 2)
        self.assertEqual(dummy._internal['y'], 2)



# -*- coding: utf-8 -*-
"""
    pimpmyclass.stats
    ~~~~~~~~~~~~~~~~~

    Implements an statistical accumulator


    Taken from Lantz 0.3

    :copyright: 2019 by Lantz Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from collections import namedtuple
from contextlib import contextmanager
from time import perf_counter


#: Data structure
class Stats(namedtuple('Stats', 'last count mean std min max')):

    def __str__(self):
        return 'Stats (hits: %d)' % self.count

    __repr__ = __str__

def stats(state):
    """Return the statistics for given state.

    Parameters
    ----------
    state : RunningState

    Returns
    -------
    Stats named tuple
    """
    if not state.count:
        return Stats(0, 0, 0, 0, 0, 0)

    mean = state.sum / state.count
    std = (float(state.sum2 - 2.0 * state.sum * mean + state.count * mean ** 2.0) / float(state.count)) ** 0.5
    return Stats(state.last, state.count,
                 mean, std, state.min, state.max)


class RunningState:
    """Accumulator for events.

    Parameters
    ----------
    value :
        first value to add.

    Returns
    -------

    """

    def __init__(self, value=None):
        if value is not None:
            self.add(value)

    def __getattr__(self, key):
        if key in ('last', 'count', 'sum', 'sum2'):
            return 0
        if key == 'min':
            return float('inf')
        if key == 'max':
            return float('-inf')
        raise AttributeError('{} is not a valid attribute of RunningState'.format(key))

    def add(self, value):
        """Add to the accumulator.

        Parameters
        ----------
        value :
            value to be added.

        Returns
        -------

        """
        self.last = value
        self.count += 1
        self.sum += value
        self.sum2 += value * value
        self.min = min(self.min, value)
        self.max = max(self.max, value)


class RunningStats(dict):
    """Accumulator for categorized event statistics.
    """

    def add(self, key, value):
        """Add an event to a given accumulator.

        Parameters
        ----------
        key :
            category to which the event should be added.
        value :
            value of the event.

        Returns
        -------

        """
        if key in self:
            super().__getitem__(key).add(value)
        else:
            super().__setitem__(key, RunningState(value))

    def stats(self, key):
        """Return the statistics for the current accumulator.

        Parameters
        ----------
        key :
            

        Returns
        -------

        """
        if key in self:
            return stats(super().__getitem__(key))
        else:
            return Stats(0, 0, 0, 0, 0, 0)

    @contextmanager
    def time(self, key):
        tic = perf_counter()
        try:
            yield
            self.add(key, perf_counter() - tic)
        except Exception as e:
            self.add('failed_' + key, perf_counter() - tic)
            raise e


# -*- coding: utf-8 -*-
"""
    pimpmyclass.dictprops
    ~~~~~~~~~~~~~~~~~~~~~

    DictProperty is a property that behaves like a dictionary.

    Wrapped get methods must have one arguments: key
    Wrapped set methods must have one arguments: key and value

    :copyright: 2019 by pimpmyclass Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import enum

from .helpers import missingdict, require, DictPropertyNameKey
from .mixins import CacheMixin, ObservableMixin
from .props import NamedProperty


class DictProperty(NamedProperty):

    _subproperties = None
    _subproperty_init = NamedProperty

    def __init__(self, *args, **kwargs):
        self.keys = kwargs.pop('keys', None)
        self._subproperties = {}
        super().__init__(*args, **kwargs)

        self._kwargs['keys'] = self.keys


    def __get__(self, instance, owner=None):

        if instance is None:
            return self

        return BoundedDictProperty(self, instance)

    def __set__(self, instance, value):

        if not isinstance(value, dict):
            raise AttributeError('The dictionary property (%s) cannot be set to a type %s '
                                 'You probably want to do something like:'
                                 'obj.prop[index] = value or obj.prop = dict-like' % (self.name, type(value)))

        for key, value in value.items():
            self.setitem(instance, key, value)

    def __delete__(self, instance):
        raise AttributeError('{} is a permanent feat of {}'.format(self.name, instance.__class__.__name__))

    def build_subproperty(self, key, fget, fset, instance=None):
        p = self._subproperty_init(
            fget=fget,
            fset=fset
        )
        return p

    def subproperty(self, instance, key):

        if key not in self._subproperties:
            p = self.build_subproperty(key,
                                       lambda s: self.fget(s, key) if self.fget is not None else None,
                                       lambda s, v: self.fset(s, key, v) if self.fset is not None else None,
                                       instance)
            assert isinstance(p, NamedProperty)
            p.__set_name__(instance.__class__, DictPropertyNameKey(self.name, key))
            self._subproperties[key] = p

        return self._subproperties[key]

    def getitem(self, instance, key):
        return self.subproperty(instance, key).__get__(instance)

    def setitem(self, instance, key, value):
        return self.subproperty(instance, key).__set__(instance, value)

    def delitem(self, instance, key):
        return self.subproperty(instance, key).__delete__(instance)

    def getall(self, instance):
        return {key: self.getitem(instance, key) for key in self._subproperties.keys()}


class BoundedDictProperty:
    """Helper class to provide indexed access to DictFeat.
    """

    def __init__(self, dictfeat, instance):
        self.instance = instance
        self.df = dictfeat

    def _get_key_val(self, key):

        keys = self.df.keys

        if keys is None:
            return key

        if isinstance(keys, enum.EnumMeta):
            if isinstance(key, enum.Enum):
                return key.value

            elif isinstance(key, str):
                try:
                    key = keys[key]
                except KeyError:
                    raise KeyError('{} is not valid key for {} {}'.format(key, self.df.name, keys))

                return key.value

            else:
                raise KeyError('{} is not valid key for {} {}'.format(key, self.df.name, keys))

        elif isinstance(keys, dict):
            try:
                key = keys[key]
            except KeyError:
                raise KeyError('{} is not valid key for {} {}'.format(key, self.df.name, keys))

        elif isinstance(keys, (set, list, tuple)):
            if key not in keys:
                raise KeyError('{} is not valid key for {} {}'.format(key, self.df.name, keys))

        return key

    def __getitem__(self, key):

        if self.df.fget is None:
            raise AttributeError('{} is a read-only feat'.format(self.df.name))

        key = self._get_key_val(key)

        return DictProperty.getitem(self.df, self.instance, key)

    def __setitem__(self, key, value):

        if self.df.fset is None:
            raise AttributeError('{} is a write-only feat'.format(self.df.name))

        key = self._get_key_val(key)

        DictProperty.setitem(self.df, self.instance, key, value)

    def __delete__(self, instance):
        if self.df.fdel:
            raise AttributeError('{} is a permanent feat of {}'.format(self.df.name, instance.__class__.__name__))

    def __repr__(self):
        return '%r.%s[]' % (self.instance, self.df.name)

    def __getattr__(self, item):
        return getattr(self.df, item)


class DictCacheProperty(DictProperty):

    def __set_name__(self, owner, name):
        require(self, owner, name, CacheMixin)

        super().__set_name__(owner, name)

    def recall(self, instance):
        grab = {key: prop.recall(instance) for key, prop in self._subproperties.items()}
        return missingdict(instance._cache_unset_value, grab)

    def store(self, instance, value):
        for name, prop in self._subproperties.items():
            prop.store(instance, value[name])

    def invalidate_cache(self, instance):
        for _, prop in self._subproperties.items():
            prop.invalidate_cache(instance)


class DictObservableProperty(DictCacheProperty):

    def __set_name__(self, owner, name):
        require(self, owner, name, CacheMixin, ObservableMixin)
        setattr(owner, name + '_changed', owner._observer_signal_init())

        super().__set_name__(owner, name)

    # Signals are emitted by subproperties.


# -*- coding: utf-8 -*-

import unittest

import statistics as stats
import random

from pimpmyclass.stats import RunningStats, stats as calc_stats


class StatsTest(unittest.TestCase):

    def test_grouping(self):
        rs = RunningStats()
        for key in ('first', 'second'):
            values = [random.random() for _ in range(20)]
            for ndx, value in enumerate(values, 1):
                rs.add(key, value)
                if ndx == 1:
                    continue
                s = rs.stats(key)

                self.assertAlmostEqual(rs[key].last, value)
                self.assertAlmostEqual(rs[key].count, ndx)
                self.assertAlmostEqual(rs[key].sum, sum(values[:ndx]))
                self.assertAlmostEqual(rs[key].sum2, sum(v ** 2 for v in values[:ndx]))
                self.assertAlmostEqual(rs[key].min, min(values[:ndx]))
                self.assertAlmostEqual(rs[key].max, max(values[:ndx]))

                self.assertAlmostEqual(s.last, value)
                self.assertAlmostEqual(s.count, ndx)
                self.assertAlmostEqual(s.mean, stats.mean(values[:ndx]))
                self.assertAlmostEqual(s.std, stats.pstdev(values[:ndx]))
                self.assertAlmostEqual(s.min, min(values[:ndx]))
                self.assertAlmostEqual(s.max, max(values[:ndx]))

    def test_wrong_attribute(self):
        rs = RunningStats()
        with self.assertRaises(AttributeError):
            out = rs.non_existing_attr

    def test_failed(self):
        rs = RunningStats()
        with rs.time('test'):
            x = 0

        with self.assertRaises(Exception):
            with rs.time('test'):
                raise Exception

        self.assertEqual(rs.stats('test').count, 1)
        self.assertEqual(rs.stats('failed_test').count, 1)

    def test_empty(self):
        rs = RunningStats()
        self.assertEqual(rs.stats('test'), (0, ) * 6)
        with rs.time('test'):
            x = 0
        self.assertNotEqual(rs.stats('test'), (0, ) * 6)
        self.assertEqual(rs.stats('other'), (0, ) * 6)

    def test_empty_calc(self):
        rs = RunningStats()
        self.assertEqual(calc_stats(rs.stats('test')), (0, ) * 6)



# -*- coding: utf-8 -*-
"""
    pimpmyclass.methods
    ~~~~~~~~~~~~~~~~~~~

    Implement pimped methods:

    - NamedMethod: has a name and can be configured via kwargs (see common.Config).
    - StorageMethod: can store and retrieve information from the instance to which is attached.
    - StatsMethod: keep stats of all calls.
    - LogMethod: logs all calls.
    - LockMethod: limits access to the instance to one thread at a time via re-entrant lock.
    - InstanceConfigurableMethod: can be configured via kwargs and those values modified in
      an instance dependent manner (see common.InstanceConfig).
    - TransformMethod: provides a way to convert input or output values.

    Some of these methods require that the class containing them derives from a particular mixin.

    :copyright: 2019 by pimpmyclass Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""


from collections import defaultdict
import functools
import inspect
import weakref

from .common import NamedCommon, Config, InstanceConfig
from .helpers import require
from .mixins import LockMixin, LogMixin, StorageMixin, BaseLogMixin
from .stats import RunningStats


class NamedMethod(NamedCommon):

    _func = None

    @property
    def name(self):
        return self._func.__name__

    @property
    def signature(self):
        return inspect.signature(self._func)

    @property
    def parameters(self):
        return tuple(self.signature.parameters.keys())

    def check_signature(self, func):
        pass

    def __call__(self, func):
        self.check_signature(func)
        self._func = func

        self.__doc__ = func.__doc__

        class NewCallable:

            def __get__(selfie, instance, owner=None):
                if instance is None:
                    return self

                func = functools.partial(self.call, instance)
                func.__wrapped__ = self._func
                return func

            def __name__(self):
                return func.__name__

            def __set_name__(selfie, owner, name):
                self.__set_name__(owner, name)

            def __getattr__(selfie, item):
                return getattr(self, item)

            def __call__(selfie, instance, *args, **kwargs):
                return self._func(instance, *args, **kwargs)

        obj = NewCallable()

        return obj

    def __newcall__(self, instance, *args, **kwargs):
        return self.call(instance, *args, **kwargs)

    def call(self, instance, *args, **kwargs):
        return self._func(instance, *args, **kwargs)

    def raw_call(self, instance, *args, **kwargs):
        return self._func(instance, *args, **kwargs)


class LockMethod(NamedMethod):

    def __set_name__(self, owner, name):
        require(self, owner, name, LockMixin)

        super().__set_name__(owner, name)

    def call(self, instance, *args, **kwargs):
        with instance.lock:
            return super().call(instance, *args, **kwargs)


class LogMethod(NamedMethod):

    log_values = Config(default=True)

    def _to_log(self, instance, value):
        if self.log_values is True:
            return value

        elif callable(self.log_values):
            try:
                return self.log_values(value)
            except Exception as e:
                instance.log_error('Could not convert value to log in %s, logging type: e', self.name, e)
                return type(value)

        return type(value)

    def _args_kwargs_to_log(self, instance, args, kwargs):
        if self.log_values is True:
            return args, kwargs

        elif callable(self.log_values):
            try:
                return tuple(self.log_values(arg) for arg in args), {k: self.log_values(v) for k, v in kwargs.items()}
            except Exception as e:
                instance.log_error('Could not convert value to log in %s, logging type: e', self.name, e)

        return tuple(type(arg) for arg in args), {k: type(v) for k, v in kwargs.items()}

    def __set_name__(self, owner, name):
        require(self, owner, name, LogMixin)

        super().__set_name__(owner, name)

    def call(self, instance, *args, **kwargs):
        if args or kwargs:
            _args, _kwargs = self._args_kwargs_to_log(instance, args, kwargs)
            instance.log_info('Calling %s with (%s, %s))', self.name, _args, _kwargs)
        else:
            instance.log_info('Calling %s', self.name)

        try:
            out = super().call(instance, *args, **kwargs)
            instance.log_info('%s returned %s', self.name, self._to_log(instance, out))
            return out
        except Exception as e:
            instance.log_error('While calling %s: %s', self.name, e)
            raise e


class StorageMethod(NamedMethod):
    """A property that can store and retrieve information in the instance
    to which is attached.

    Methods and descriptors are class attributes and therefore any attempt to
    naively modify one of their attributes for a single instance of the parent
    class will propagate to all instances. This property overcomes this problem
    by storing information at the instance level.

    The information is stored in uniquely a specified namespace defined by
    the derived class. Inside that storage, another namespace is specified
    using the property name.

    Derived class should use the dynamically created _store_get and _store_set
    to retrieve and store information.

    ..note: Derived classes must override the following variables:

        _storage_ns : str
            Defines a unique namespace under which the information of the
            derived class is stored.

        _storage_ns_init : callable
            Called upon initialization of the storage to initialize the
            specific storage of the namespace.


    **Requires** that the owner class inherits :class:`pimpmyclass.mixins.StorageMixin`.
    """

    # Stores namespace to StorageProperty subclass
    # It cannot be dunder because it is accessed by __init_subclass__
    _storage_sub_ns_cls = weakref.WeakValueDictionary()

    _storage_ns = ''
    _storage_ns_init = None

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        ns = cls._storage_ns
        if not ns:
            raise ValueError('Class %s must specify a storage namespace '
                             ' as required by StorageMethod' % cls)

        if ns in cls._storage_sub_ns_cls:
            if not issubclass(cls, cls._storage_sub_ns_cls[ns]):
                raise ValueError('Class %r storage namespace (%s) collides with '
                                 'class %r' % (cls, ns, cls._storage_sub_ns_cls[ns]))
        else:
            if cls._storage_ns_init is None:
                raise ValueError('Class %s must specify a storage initializer '
                                 ' as required by StorageMethod' % cls)

            cls._storage_sub_ns_cls[ns] = cls

            # Create partial versions of _store_get and _store_get
            #  with the corresponding namespace
            #  and store it in the specific subclass.
            cls._store_get = functools.partial(cls._ns_store_get, namespace=cls._storage_ns)
            cls._store_set = functools.partial(cls._ns_store_set, namespace=cls._storage_ns)
            cls._store_del = functools.partial(cls._ns_store_del, namespace=cls._storage_ns)

    def __set_name__(self, owner, name):
        require(self, owner, name, StorageMixin)

        super().__set_name__(owner, name)

    def _store_get(self, instance):
        return self._ns_store_get(instance, self._storage_ns)

    def _store_set(self, instance, value):
        return self._ns_store_set(instance, value, self._storage_ns)

    def _store_del(self, instance):
        return self._ns_store_del(instance, self._storage_ns)

    def _ns_store_get(self, instance, namespace):
        sto = instance.storage

        if namespace not in sto:
            cls = self._storage_sub_ns_cls[namespace]
            sto[namespace] = cls._storage_ns_init(instance)

        return sto[namespace][self.name]

    def _ns_store_set(self, instance, value, namespace):
        sto = instance.storage

        if namespace not in sto:
            cls = self._storage_sub_ns_cls[namespace]
            sto[namespace] = cls._storage_ns_init(instance)

        sto[namespace][self.name] = value

    def _ns_store_del(self, instance, namespace):
        sto = instance.storage

        if namespace not in sto:
            cls = self._storage_sub_ns_cls[namespace]
            sto[namespace] = cls._storage_ns_init(instance)

        del sto[namespace][self.name]


class StatsMethod(StorageMethod):
    """A property that keep stats on get and set calls.

    Stats can be retrieved with the `stat` methods and
    the following keys:
    - call
    - failed_call

    The following statistics are provided in a namedtuple:

        last : float
            most recent duration (seconds).
        count : int
            number of operations.
        mean : float
            average duration per operation (seconds).
        std : float
            standard deviation of the duration (seconds).
        min : float
            shortest duration (seconds).
        max : float
            longest duration (seconds).


    **Requires** that the owner class inherits :class:`pimpmyclass.mixins.StorageMixin`.
    """

    _storage_ns = 'statsm'
    _storage_ns_init = lambda _: defaultdict(RunningStats)

    def call(self, instance, *args, **kwargs):
        with StatsMethod._store_get(self, instance).time('call'):
            return super().call(instance, *args, **kwargs)

    def stats(self, instance, key):
        return StatsMethod._store_get(self, instance).stats(key)


class InstanceConfigurableMethod(StorageMethod):

    _storage_ns = 'iconfigm'
    _storage_ns_init = lambda _: defaultdict(dict)

    def config_get(self, instance, key):

        if instance is None:
            return super().config_get(None, key)

        try:
            return InstanceConfigurableMethod._store_get(self, instance)[key]
        except KeyError:
            return super().config_get(None, key)

    def config_set(self, instance, key, value):
        if instance is None:
            super().config_set(None, key, value)
        else:
            InstanceConfigurableMethod._store_get(self, instance)[key] = value

        self.on_config_set(instance, key, value)


class TransformMethod(InstanceConfigurableMethod):

    _transformations = None
    _storage_ns = 'transformationsm'
    _storage_ns_init = lambda _: defaultdict(dict)

    params = InstanceConfig(default=None)

    def __set_name__(self, owner, name):
        require(self, owner, name, StorageMixin, BaseLogMixin)

        super().__set_name__(owner, name)

        sig = self.signature
        param_names = self.parameters
        if self.params:
            if not isinstance(self.params, dict):
                if len(sig.parameters) > 2:
                    raise ValueError('The syntax for methods with multiple arguments '
                                     'in the constructor has been deprecated. Use the '
                                     'cleaner "param" syntax in %s', name)

                last = param_names[-1]
                self.params = {last: self.params}
        else:
            self.params = {}
            try:
                self.params = dict(self._func.__transform_params__)
                for k in self.params.keys():
                    if k == '<ret>':
                        continue
                    if k not in param_names:
                        raise ValueError('%s is not an argument name of %s', k, name)
                del self._func.__transform_params__
            except AttributeError:
                pass

        assert isinstance(self.params, dict)

    def check_signature(self, func):
        super().check_signature(func)

        if not self.params:
            return

        if not isinstance(self.params, dict):
            return

        names = tuple(inspect.signature(func).parameters.keys())

        p = self.params.pop(None, None)

        if not p:
            return

        for name in names[1:]:
            if name not in self.params:
                self.params[name] = p

    def call(self, instance, *args, **kwargs):

        sig = self.signature
        ba = sig.bind(instance, *args, **kwargs)

        t_arg = self.params_iget(instance)
        if t_arg:
            r_arg = t_arg.pop('<ret>', None)
        else:
            r_arg = None
        new_ba = {}

        noop = lambda x: x

        if t_arg:
            
            for param in sig.parameters.values():
                if param.kind != param.POSITIONAL_OR_KEYWORD:
                    raise ValueError('Only named POSITIONAL_OR_KEYWORD arguments are currently '
                                     'supported by transformations.')

            for k, v in ba.arguments.items():
                new_ba[k] = t_arg.get(k, noop)(v)

            ba = sig.bind(**new_ba)
            instance.log_info('<T> Calling %s with (%s, %s)', self.name, ba.args, ba.kwargs)
            try:
                out = super().call(*ba.args, **ba.kwargs)
            except Exception as e:
                instance.log_error('While pre-processing (%s, %s) for %s: %s', ba.args, ba.kwargs, self.name, e)
                raise e
        else:
            out = super().call(*ba.args, **ba.kwargs)

        if r_arg:
            try:
                out = r_arg(out)
            except Exception as e:
                instance.log_error('While post-processing %s for %s: %s', out, self.name, e)
                raise e

        return out

    @classmethod
    def param(cls, names, func):
        """Add modifiers to a specific parameter.

        See Action for more information.
        """
        def decorator(f):
            _param_memo(f, names, func)
            return f
        return decorator

    @classmethod
    def ret(cls, func):
        """Add modifiers to the return value.

        See Action for more information.
        """
        def decorator(f):
            _param_memo(f, '<ret>', func)
            return f
        return decorator


def _param_memo(obj, name, func):
    """Helper function to allow composition of modifiers in a single action.
    """
    if not isinstance(name, str):
        for n in name:
            _param_memo(obj, n, func)
        return

    if not hasattr(obj, '__transform_params__'):
        obj.__transform_params__ = []
    obj.__transform_params__.append((name, func))



# -*- coding: utf-8 -*-
"""
    pimpmyclass.commoon
    ~~~~~~~~~~~~~~~~~~~

    Provides Config, InstanceConfig and a base class for NamedProperties and Methods

    :copyright: 2019 by pimpmyclass Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

import inspect

from .helpers import append_lines_to_docstring, require_any


#: This sentinel indicates a configuration value of a property has not been set
#: It is only for internal use as all configurations must have a defined value
#: either by a kwargs during construction or because it has a default value.
CONFIG_UNSET = object()


class Config:

    def __init__(self, valid_values=(), valid_types=(), check_func=None, default=CONFIG_UNSET, doc=''):
        self.valid_values = valid_values
        self.valid_types = valid_types
        self.check_func = check_func
        self.default = default
        self.__doc__ = doc

    def _check_value(self, value, name):
        if self.valid_values and value not in self.valid_values:
            raise ValueError('%r is not a valid value for %s. Should be in %r' %
                             (value, name, self.valid_values))

        if self.valid_types and not isinstance(value, self.valid_types):
            raise TypeError('%r is not a valid type for %s. Should be in %r' %
                            (value, name, self.valid_types))

        if self.check_func:
            try:
                ok = self.check_func(value)
            except Exception as e:
                raise ValueError('The value provided for %s does not pass the check function: %s' % (name, e))
            if not ok:
                raise ValueError('The value provided for %s does not pass the check function')

    def _common(self, owner, name):

        if owner._config_objects is None:
            owner._config_objects = {name: self}
        else:
            owner._config_objects[name] = self

        def _get(selfie):
            return selfie.config_get(None, name)

        def _set(selfie, value):
            self._check_value(value, name)
            return selfie.config_set(None, name, value)

        setattr(owner, name, property(_get, _set))

    def __set_name__(self, owner, name):
        from .props import NamedProperty
        from .methods import NamedMethod
        require_any(self, owner, name, NamedProperty, NamedMethod)

        self._common(owner, name)


class InstanceConfig(Config):

    def __set_name__(self, owner, name):
        # from .props import InstanceConfigurableProperty
        # from .methods import InstanceConfigurableMethod
        # require_any(self, owner, name, InstanceConfigurableProperty, InstanceConfigurableMethod)

        self._common(owner, name)

        def _iget(selfie, instance):
            return selfie.config_get(instance, name)

        def _iset(selfie, instance, value):
            self._check_value(value, name)
            return selfie.config_set(instance, name, value)

        setattr(owner, name + '_iget', _iget)
        setattr(owner, name + '_iset', _iset)


class MetaDoc(type):

    def __new__(cls, name, bases, attrs):
        attrs['_doc'] = attrs.get('__doc__', '')
        return super(MetaDoc, cls).__new__(cls, name, bases, attrs)

    @property
    def __doc__(self):
        return self.fulldoc(self._doc)


class NamedCommon(metaclass=MetaDoc):

    _name = ''
    _kwargs = None

    _config = None
    _config_objects = None

    def __init__(self, **kwargs):

        self._kwargs = {}

        self._config = {}
        for base_class in inspect.getmro(self.__class__):
            if getattr(base_class, '_config_objects', None):
                self._config.update({name: obj.default for name, obj in base_class._config_objects.items()})

        for k in self._config.keys():
            if k in kwargs:
                v = kwargs.pop(k)
                setattr(self, k, v)
                self._kwargs[k] = v

        if kwargs:
            raise TypeError("%s() got an unexpected keyword argument '%s'" %
                            (self.__class__.__name__, list(kwargs.keys())[0]))

        if self._config:
            missing = tuple(k for k, v in self._config.items() if v is CONFIG_UNSET)
            if missing:
                raise TypeError("%s() is missing %d positional argument%s: %s" %
                                (self.__class__.__name__, len(missing),
                                 's' if len(missing) > 1 else '', ','.join(missing)))

    def __set_name__(self, owner, name):
        self._name = name

    @classmethod
    def fulldoc(cls, doc):

        if not cls._config_objects:
            return doc

        doc = doc or ''

        lines = ['',
                 '',
                 'Other parameters',
                 '----------------']

        for name, obj in cls._config_objects.items():
            desc = []
            if obj.valid_values:
                desc.append(' or '.join(repr(el) for el in obj.valid_values))
            if obj.valid_types:
                desc.append(' or '.join(el.__name__ for el in obj.valid_types))
            if obj.check_func:
                desc.append(' Note: checking function')

            if desc:
                desc = ' and '.join(desc)
            else:
                desc = ''
            if obj.default is not CONFIG_UNSET:
                if desc:
                    desc += ' '
                desc += '(default=%r)' % obj.default

            if desc:
                lines.append('%s : %s' % (name, desc))
            else:
                lines.append(name)

            if obj.__doc__:
                lines.append('    ' + obj.__doc__)

        lines.append('')

        return append_lines_to_docstring(lines, doc, mixed_fallback='')

    def config_get(self, instance, key):
        return self._config[key]

    def config_set(self, instance, key, value):
        self._config[key] = value

    def config_iter(self, instance):
        for key in self._config.keys():
            yield key, self.config_get(instance, key)

    def on_config_set(self, instance, key, value):
        pass



# -*- coding: utf-8 -*-
"""
    pimpmyclass.props
    ~~~~~~~~~~~~~~~~~

    Implement pimped properties:

    - NamedProperty: has a name and can be configured via kwargs (see common.Config).
    - StorageProperty: can store and retrieve information from the instance to which is attached.
    - StatsProperty: keep stats of all get and set operations.
    - LogProperty: logs all get and set operations.
    - LockProperty: limits access to the instance to one thread at a time via re-entrant lock.
    - InstanceConfigurableProperty: can be configured via kwargs and those values modified in
      an instance dependent manner (see common.InstanceConfig).
    - TransformProperty: provides a way to convert get or set values.
    - CacheProperty: has a cache that can be invalidated.
    - GetCacheProperty: stores in the cache the output of the last get operation.
    - SetCacheProperty: stores in the cache the input to the last set operation.
    - PreventUnnecessarySetProperty: do not set the value if it matches the value in the cache.
    - ReadOnceProperty: do not get the value if it is already in the cache.
    - ObservableProperty: emits a signal if the cache changes.

    Some of these properties require that the class containing them derives from a particular mixin.

    :copyright: 2019 by pimpmyclass Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from collections import defaultdict
import functools as ft
import inspect
import weakref

from .common import NamedCommon, Config, InstanceConfig
from .helpers import missingdict, require, DictPropertyNameKey
from .stats import RunningStats
from .mixins import StorageMixin, BaseLogMixin, LockMixin, CacheMixin, ObservableMixin


class NamedProperty(NamedCommon):
    """A property that takes the name of the class attribute to which it is assigned.

    Accepts instance of :class:`pimpmyclass.Config` as configuration values that
    automatically get populated from kwargs.
    """

    def __init__(self, fget=None, fset=None, fdel=None, doc=None, **kwargs):
        super().__init__(**kwargs)
        self.fget = fget
        self.fset = fset
        self.fdel = fdel

        if doc is None and fget is not None:
            self.__doc__ = fget.__doc__
        elif doc is not None:
            self.__doc__ = doc

    @property
    def name(self):
        return self._name

    def get_notify(self, instance, value):
        pass

    def set_notify(self, instance, value):
        pass

    def __call__(self, func):
        if self.fget is None:
            return self.getter(func)

        return self.setter(func)

    def __get__(self, instance, objtype=None):
        if instance is None:
            return self

        if self.fget is None:
            raise AttributeError('%s is a read-only property of %s' %
                                 (self.name, instance.__class__.__name__))

        value = self.get(instance, objtype)
        self.get_notify(instance, value)
        return value

    def __set__(self, instance, value):

        if self.fset is None:
            raise AttributeError('%s is a write-only property of %s' %
                                 (self.name, instance.__class__.__name__))

        self.set(instance, value)
        self.set_notify(instance, value)

    def __delete__(self, instance):

        if self.fdel is None:
            raise AttributeError('%s is a permanent feat of %s' %
                                 (self.name, instance.__class__.__name__))

        self.delete(instance)

    @property
    def fget_signature(self):
        return inspect.signature(self.fget)

    def get(self, instance, objtype):
        return self.fget(instance)

    def set(self, instance, value):
        return self.fset(instance, value)

    def delete(self, instance):
        return self.fdel(instance)

    def getter(self, fget):
        return type(self)(fget, self.fset, self.fdel, self.__doc__,
                          **getattr(self, '_kwargs') or {})

    def setter(self, fset):
        return type(self)(self.fget, fset, self.fdel, self.__doc__,
                          **getattr(self, '_kwargs') or {})

    def deleter(self, fdel):
        return type(self)(self.fget, self.fset, fdel, self.__doc__,
                          **getattr(self, '_kwargs') or {})


class StorageProperty(NamedProperty):
    """A property that can store and retrieve information in the instance
    to which is attached.

    **Requires** that the owner class inherits :class:`pimpmyclass.mixins.StorageMixin`.

    Notes
    -----

    The information is stored in uniquely a specified namespace defined by
    the derived class. Inside that storage, another namespace is specified
    using the property name.

    Derived class should use the dynamically created _store_get and _store_set
    to retrieve and store information.

    .. note:: Derived classes must override the following variables:

        _storage_ns : str
            Defines a unique namespace under which the information of the
            derived class is stored.

        _storage_ns_init : callable
            Called upon initialization of the storage to initialize the
            specific storage of the namespace.

    """

    # Stores namespace to StorageProperty subclass
    # It cannot be dunder because it is accessed by __init_subclass__
    _storage_sub_ns_cls = weakref.WeakValueDictionary()

    _storage_ns = ''
    _storage_ns_init = None

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)

        ns = cls._storage_ns
        if not ns:
            raise ValueError('Class %s must specify a storage namespace '
                             ' as required by StorageProperty' % cls)

        if ns in cls._storage_sub_ns_cls:
            if not issubclass(cls, cls._storage_sub_ns_cls[ns]):
                if cls is not cls._storage_sub_ns_cls[ns]:
                    raise ValueError('Class %r storage namespace (%s) collides with '
                                     'class %r' % (cls, ns, cls._storage_sub_ns_cls[ns]))
        else:
            if cls._storage_ns_init is None:
                raise ValueError('Class %s must specify a storage initializer '
                                 ' as required by StorageProperty' % cls)

            cls._storage_sub_ns_cls[ns] = cls

            # Create partial versions of _store_get and _store_get
            #  with the corresponding namespace
            #  and store it in the specific subclass.
            cls._store_get = ft.partial(cls._ns_store_get, namespace=cls._storage_ns)
            cls._store_set = ft.partial(cls._ns_store_set, namespace=cls._storage_ns)
            cls._store_del = ft.partial(cls._ns_store_del, namespace=cls._storage_ns)

    def __set_name__(self, owner, name):
        require(self, owner, name, StorageMixin)

        super().__set_name__(owner, name)

    def _store_get(self, instance):
        return self._ns_store_get(instance, self._storage_ns)

    def _store_set(self, instance, value):
        return self._ns_store_set(instance, value, self._storage_ns)

    def _store_del(self, instance):
        return self._ns_store_del(instance, self._storage_ns)

    def _ns_store_get(self, instance, namespace):
        sto = instance.storage

        if namespace not in sto:
            cls = self._storage_sub_ns_cls[namespace]
            sto[namespace] = cls._storage_ns_init(instance)

        return sto[namespace][self.name]

    def _ns_store_set(self, instance, value, namespace):
        sto = instance.storage

        if namespace not in sto:
            cls = self._storage_sub_ns_cls[namespace]
            sto[namespace] = cls._storage_ns_init(instance)

        sto[namespace][self.name] = value

    def _ns_store_del(self, instance, namespace):
        sto = instance.storage

        if namespace not in sto:
            cls = self._storage_sub_ns_cls[namespace]
            sto[namespace] = cls._storage_ns_init(instance)

        try:
            del sto[namespace][self.name]
        except KeyError:
            pass


class StatsProperty(StorageProperty):
    """A property that keep stats on get and set calls.

    Stats can be retrieved with the `stat` methods and
    the following keys:

    - get
    - set
    - failed_get
    - failed_set

    The following statistics are provided in a namedtuple:

        last : float
            most recent duration (seconds).
        count : int
            number of operations.
        mean : float
            average duration per operation (seconds).
        std : float
            standard deviation of the duration (seconds).
        min : float
            shortest duration (seconds).
        max : float
            longest duration (seconds).


    **Requires** that the owner class inherits :class:`pimpmyclass.mixins.StorageMixin`.
    """

    _storage_ns = 'stats'
    _storage_ns_init = lambda _: defaultdict(RunningStats)

    def get(self, instance, objtype):

        with StatsProperty._store_get(self, instance).time('get'):
            return super().get(instance, objtype)

    def set(self, instance, value):

        with StatsProperty._store_get(self, instance).time('set'):
            return super().set(instance, value)

    def stats(self, instance, key):
        return StatsProperty._store_get(self, instance).stats(key)


class LogProperty(NamedProperty):
    """A property that log operations.

    **Requires** that the owner class inherits :class:`pimpmyclass.mixins.LogMixin`.
    """

    log_values = Config(default=True,
                        doc='If False, just the types (not the values) will logged.\n'
                            'An obj to str callable can be provided to perform custom serialization.' )

    def __set_name__(self, owner, name):
        require(self, owner, name, BaseLogMixin)

        super().__set_name__(owner, name)

    def _to_log(self, instance, value):
        if self.log_values is True:
            return value

        elif callable(self.log_values):
            try:
                return self.log_values(value)
            except Exception as e:
                instance.log_error('Could not convert value to log in %s, logging type: e', self.name, e)

        return type(value)

    def get(self, instance, objtype):

        instance.log_info('Getting %s', self.name)
        try:
            value = super().get(instance, objtype)
            log_value = self._to_log(instance, value)
            instance.log_debug('Got %s for %s', log_value, self.name)
        except Exception as e:
            instance.log_error('While getting %s: %s', self.name, e)
            raise e

        return value

    def set(self, instance, value):
        log_value = self._to_log(instance, value)
        instance.log_debug('Setting %s to %s', self.name, log_value)
        try:
            super().set(instance, value)
            instance.log_debug('%s was set to %s', self.name, log_value)
        except Exception as e:
            instance.log_error('While setting %s to %s: %s', self.name, log_value, e)
            raise e


class LockProperty(NamedProperty):
    """A property that with a set or get Lock.

    **Requires** that the owner class inherits :class:`pimpmyclass.mixins.LogMixin`.
    """

    def __set_name__(self, owner, name):
        require(self, owner, name, LockMixin)

        super().__set_name__(owner, name)

    def get(self, instance, objtype):

        with instance.lock:
            return super().get(instance, objtype)

    def set(self, instance, value):
        with instance.lock:
            return super().set(instance, value)


class InstanceConfigurableProperty(StorageProperty):
    """A property that contains owner instance specific configuration variable.

    **Requires** that the owner class inherits :class:`pimpmyclass.mixins.StorageMixin`.
    """

    _storage_ns = 'iconfig'
    _storage_ns_init = lambda _: defaultdict(dict)

    def config_get(self, instance, key):

        if instance is None:
            return super().config_get(None, key)

        try:
            return InstanceConfigurableProperty._store_get(self, instance)[key]
        except KeyError:
            return super().config_get(None, key)

    def config_set(self, instance, key, value):
        if instance is None:
            super().config_set(None, key, value)
        else:
            InstanceConfigurableProperty._store_get(self, instance)[key] = value

        self.on_config_set(instance, key, value)


class TransformProperty(InstanceConfigurableProperty):
    """A property that can transform value before a set operation
    or after a get operation.

    **Requires** that the owner class inherits :class:`pimpmyclass.mixins.InstanceConfigurableProperty`.
    """

    pre_set = InstanceConfig(default=None, check_func=lambda x: x is None or callable(x))
    post_get = InstanceConfig(default=None, check_func=lambda x: x is None or callable(x))

    def __set_name__(self, owner, name):
        require(self, owner, name, StorageMixin, BaseLogMixin)

        super().__set_name__(owner, name)

    def get(self, instance, objtype):

        value = super().get(instance, objtype)

        transform = self.post_get_iget(instance)

        if not transform:
            return value

        try:
            value = transform(value)
            instance.log_debug('<T> Got %s for %s', value, self.name)
            return value
        except Exception as e:
            instance.log_error('While post-processing %s for %s: %s', value, self.name, e)
            raise e

    def set(self, instance, value):

        transform = self.pre_set_iget(instance)

        if transform:

            try:
                value = transform(value)
                instance.log_info('<T> Setting %s = %s', self.name, value)
            except Exception as e:
                instance.log_error('While pre-processing %s for %s: %s', value, self.name, e)
                raise e

        super().set(instance, value)


class CacheProperty(StorageProperty):
    """A property that can store, recall or invalidate a cache.
    """

    _storage_ns = 'cache'
    _storage_ns_init = lambda instance: missingdict(instance._cache_unset_value)

    def __set_name__(self, owner, name):
        require(self, owner, name, CacheMixin, BaseLogMixin)

        super().__set_name__(owner, name)

    def recall(self, instance):
        return CacheProperty._store_get(self, instance)

    def store(self, instance, value):
        CacheProperty._store_set(self, instance, value)

    def invalidate_cache(self, instance):
        CacheProperty._store_del(self, instance)


class GetCacheProperty(CacheProperty):
    """A property that stores the get value in the cache.

    **Requires** that the owner class inherits :class:`pimpmyclass.mixins.StorageMixin`.
    """

    def get_notify(self, instance, value):
        super().get_notify(instance, value)

        self.store(instance, value)


class SetCacheProperty(CacheProperty):
    """A property that stores the set value in the cache.

    **Requires** that the owner class inherits :class:`pimpmyclass.mixins.StorageMixin`.
    """

    def set_notify(self, instance, value):
        super().set_notify(instance, value)

        self.store(instance, value)


class GetSetCacheProperty(GetCacheProperty, SetCacheProperty):
    """A property that stores the get or set value in the cache.

    **Requires** that the owner class inherits :class:`pimpmyclass.mixins.StorageMixin`.
    """


class PreventUnnecessarySetProperty(SetCacheProperty):
    """A property that prevents unnecessary set operations by comparing
    the value in the cache with the value to be set.

    **Requires** that the owner class inherits :class:`pimpmyclass.mixins.CacheMixin` and :class:`pimpmyclass.mixins.LogMixin`.
    """

    def set(self, instance, value):
        current_value = self.recall(instance)

        if value == current_value:
            instance.log_info('No need to set %s = %s (current=%s)', self.name, value, current_value)
            return

        super().set(instance, value)

    def force_set(self, instance, value):
        self.invalidate_cache(instance)
        self.set(instance, value)


class ReadOnceProperty(InstanceConfigurableProperty, GetCacheProperty):
    """Avoids calling the getter if the value is already in the cache.

    """

    read_once = InstanceConfig(default=False, valid_types=(bool, ))

    def get(self, instance, owner=None):
        if self.read_once_iget(instance) and self.recall(instance) is not instance._cache_unset_value:
            return self.recall(instance)

        return super().get(instance, owner)


class ObservableProperty(CacheProperty):
    """A property that emits a signal when the cached value is changed
    (either via set or get)

    **Requires** that the owner class inherits :class:`pimpmyclass.mixins.CacheMixin` and :class:`pimpmyclass.mixins.ObservableMixin`.
    """

    def __set_name__(self, owner, name):
        if isinstance(name, str):
            require(self, owner, name, CacheMixin, ObservableMixin)
            setattr(owner, name + '_changed', owner._observer_signal_init())

        super().__set_name__(owner, name)

    def store(self, instance, value):
        old_value = self.recall(instance)
        super().store(instance, value)
        if old_value != value:
            if isinstance(self.name, DictPropertyNameKey):
                getattr(instance, self.name.name + '_changed').emit(value, old_value, self.name.key)
            else:
                getattr(instance, self.name + '_changed').emit(value, old_value)



# -*- coding: utf-8 -*-

import os
import unittest


def testsuite():
    """A testsuite that has all the pimpmyclass tests.
    """
    return unittest.TestLoader().discover(os.path.dirname(__file__))


def main():
    """Runs the testsuite as command line application.
    """
    try:
        unittest.main()
    except Exception as e:
        print('Error: %s' % e)


def run():
    """Run all tests.

    :return: a :class:`unittest.TestResult` object
    """
    test_runner = unittest.TextTestRunner()
    return test_runner.run(testsuite())



# -*- coding: utf-8 -*-
"""
    pimpmyclass.helper
    ~~~~~~~~~~~~~~~~~~

    General helper functions and classes.

    :copyright: 2019 by pimpmyclass Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from collections import UserDict, namedtuple


class DictPropertyNameKey(namedtuple('DictPropertyNameKey', 'name key')):

    def __str__(self):
        return '%s[%r]' % (self.name, self.key)


class missingdict(UserDict):
    """Dictionary that returns UNSET
    """

    __unset_value = None

    def __init__(self, unset_value, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.__unset_value = unset_value

    def __missing__(self, key):
        return self.__unset_value


def keep_if_not(_skip=None, **kwargs):
    return {k: v for k, v in kwargs.items() if v is not _skip}


def require(inst, owner, name, *parents):
    """Raises a (nicely explained) type error if owner is not a subclass of all parents.

    It is useful when making sure that a certain type of property is only used within
    certain classes.

    Parameters
    ----------
    inst
        instance of a property object or method.
    owner : type
        class of an object.
    name : str
        name of the attribute to which the property object is assigned.
    parents : iterable of type
        classes that owner must inherit.

    Raises
    ------
    TypeError
        If owner does not inherit from all parents.

    """
    for parent in parents:
        if not issubclass(owner, parent):
            raise TypeError('%r is a %r but %r is not a subclass from %r as required' %
                            (name, inst.__class__.__name__, owner.__name__, parent.__name__))


def require_any(inst, owner, name, *parents):
    """Raises a (nicely explained) type error if owner is not a subclass of at least one parent.

    It is useful when making sure that a certain type of property is only used within
    certain classes.

    Parameters
    ----------
    inst
        instance of a property object.
    owner : type
        class of an object.
    name : str
        name of the attribute to which the property object is assigned.
    parents : iterable of type
        classes that owner must inherit at least one.

    Raises
    ------
    TypeError
        If owner does not inherit at least from one parents.

    """
    for parent in parents:
        if issubclass(owner, parent):
            break
    else:
        raise TypeError('%r is a %r but %r is not a subclass from any of %r as required' %
                        (name, inst.__class__.__name__, owner.__name__, parents))


def guess_indent(text, skip_first=False):
    if not text:
        return ''

    # Find the indentation of this docstring
    # by looking at the minimum number of leading chars
    # in non empty lines

    leading_spaces = 10000000000000
    leading_tabs = 10000000000000

    first_space = False
    first_tab = False

    lines = text.split('\n')

    if skip_first:
        if len(lines) == 1:
            return ''
        lines = lines[1:]

    for line in lines:
        # skip empty lines
        if not line.strip():
            continue
        leading_spaces = min(leading_spaces, len(line) - len(line.lstrip(' ')))
        leading_tabs = min(leading_tabs, len(line) - len(line.lstrip('\t')))

        first_space = first_space or line[0] == ' '
        first_tab = first_tab or line[0] == '\t'

    if first_space and first_tab:
        raise ValueError('Mixed tabs and spaces in docstring\n' + text)
    elif first_space:
        return ' ' * leading_spaces
    elif first_tab:
        return '\t' * leading_tabs
    else:
        return ''


def prepend_to_docstring(s, docstring, mixed_fallback=None):
    if not docstring:
        return s

    try:
        indent = guess_indent(docstring, False)
    except ValueError:
        if mixed_fallback is None:
            raise
        indent = mixed_fallback

    return indent + s + docstring


def append_lines_to_docstring(lines, docstring, mixed_fallback=None):
    if not docstring:
        return '\n'.join(lines)

    try:
        indent = guess_indent(docstring, True)
    except ValueError:
        if mixed_fallback is None:
            raise
        indent = mixed_fallback

    out = docstring

    for line in lines:
        if line:
            out += indent + line + '\n'
        else:
            out += '\n'
    return out + indent




# -*- coding: utf-8 -*-

import unittest
import logging

from pimpmyclass import mixins, props, helpers, common


class MemHandler(logging.Handler):

    def __init__(self):
        super().__init__()
        self.setFormatter(logging.Formatter(style='{'))
        self.history = list()

    def emit(self, record):
        self.history.append(self.format(record))


def define(proptype, *bases):

    class Dummy(*bases):

        logger_name = 'testing123'

        _internal = 3
        _prop_gs = 8

        @proptype()
        def prop(self):
            return self._internal

        @prop.setter
        def prop(self, value):
            if value is None:
                raise Exception('Arrrg!')

        @proptype()
        def prop_gs(self):
            return self._prop_gs

        @prop_gs.setter
        def prop_gs(self, value):
            self._prop_gs = value

        @proptype()
        def properr(self):
            raise Exception('GetArrrg!')


    return Dummy


class TestNamedProperty(unittest.TestCase):

    def test_name(self):

        Dummy = define(props.NamedProperty)
        self.assertEqual(Dummy.prop.name, 'prop')

    def test_readonly(self):

        class C:

            @props.NamedProperty()
            def prop(self):
                return 4

        o = C()

        self.assertEqual(o.prop, 4)

        with self.assertRaises(AttributeError):
            o.prop = 5

        with self.assertRaises(AttributeError):
            del o.prop

    def test_writeonly(self):

        class C:

            prop = props.NamedProperty()

            _value = 0

            @prop.setter
            def prop(self, value):
                self._value = value

        o = C()

        self.assertEqual(o._value, 0)
        o._value = 4
        self.assertEqual(o._value, 4)

        with self.assertRaises(AttributeError):
            o.prop

        with self.assertRaises(AttributeError):
            del o.prop

    def test_deleteonly(self):

        class C:

            prop = props.NamedProperty()

            _value = 4

            @prop.deleter
            def prop(self):
                self._value = 0

        o = C()

        self.assertEqual(o._value, 4)
        del o.prop
        self.assertEqual(o._value, 0)

        with self.assertRaises(AttributeError):
            o.prop = 4

        with self.assertRaises(AttributeError):
            o.prop

    def test_readwrite_call(self):

        class C:

            _value = 0

            @props.NamedProperty()
            def prop(self):
                return self._value

            @prop
            def prop(self, value):
                self._value = value

        o = C()

        self.assertEqual(o.prop, 0)
        o.prop = 4
        self.assertEqual(o._value, 4)

        with self.assertRaises(AttributeError):
            del o.prop


class TestOtherProperties(unittest.TestCase):

    def test_timing(self):

        with self.assertRaises(Exception):
            define(props.StatsProperty)

        Dummy = define(props.StatsProperty, mixins.StorageMixin)
        x = Dummy()
        y = Dummy()

        def g(i):
            return getattr(i.__class__, 'prop')

        self.assertIsNot(g(x).stats, g(y).stats)

        s = g(x).stats(x, 'get')
        self.assertEqual(s.count, 0)

        self.assertEqual(x.prop, 3)

        s = g(x).stats(x, 'get')
        self.assertEqual(s.count, 1)

        s = g(y).stats(y, 'get')
        self.assertEqual(s.count, 0)

        x.prop = 0
        s = g(x).stats(x, 'set')
        self.assertEqual(s.count, 1)
        s = g(y).stats(y, 'set')
        self.assertEqual(s.count, 0)

        with self.assertRaises(Exception):
            x.prop = None

        s = g(x).stats(x, 'failed_set')
        self.assertEqual(s.count, 1)

    def test_log(self):

        with self.assertRaises(Exception):
            define(props.LogProperty)

        Dummy = define(props.LogProperty, mixins.LogMixin)
        x = Dummy()

        hdl = MemHandler()
        x.logger.addHandler(hdl)
        x.logger.setLevel(logging.DEBUG)

        x.prop = 1
        y = x.prop
        with self.assertRaises(Exception):
            x.prop = None

        with self.assertRaises(Exception):
            x.properr

        self.assertEqual(hdl.history, ['Setting prop to 1',
                                       'prop was set to 1',
                                       'Getting prop',
                                       'Got 3 for prop',
                                       'Setting prop to None',
                                       'While setting prop to None: Arrrg!',
                                       'Getting properr',
                                       'While getting properr: GetArrrg!'])

    def test_log_config_false(self):

        Dummy = define(lambda: props.LogProperty(log_values=False), mixins.LogMixin)
        x = Dummy()

        hdl = MemHandler()
        x.logger.addHandler(hdl)
        x.logger.setLevel(logging.DEBUG)

        x.prop = 1
        y = x.prop
        with self.assertRaises(Exception):
            x.prop = None

        with self.assertRaises(Exception):
            x.properr

        self.assertEqual(hdl.history, ["Setting prop to <class 'int'>",
                                       "prop was set to <class 'int'>",
                                       "Getting prop",
                                       "Got <class 'int'> for prop",
                                       "Setting prop to <class 'NoneType'>",
                                       "While setting prop to <class 'NoneType'>: Arrrg!",
                                       "Getting properr",
                                       "While getting properr: GetArrrg!"])

    def test_derive_log_config_false(self):

        class MyProp(props.LogProperty):
            pass

        Dummy = define(lambda: MyProp(), mixins.LogMixin)
        x = Dummy()

        hdl = MemHandler()
        x.logger.addHandler(hdl)
        x.logger.setLevel(logging.DEBUG)

        x.prop = 1
        y = x.prop
        with self.assertRaises(Exception):
            x.prop = None

        with self.assertRaises(Exception):
            x.properr

        self.assertEqual(hdl.history, ['Setting prop to 1',
                                       'prop was set to 1',
                                       'Getting prop',
                                       'Got 3 for prop',
                                       'Setting prop to None',
                                       'While setting prop to None: Arrrg!',
                                       'Getting properr',
                                       'While getting properr: GetArrrg!'])

    def test_derive_log_config_false(self):

        class MyProp(props.TransformProperty, props.LogProperty):
            pass

        Dummy = define(MyProp, mixins.LogMixin, mixins.StorageMixin)
        x = Dummy()

        hdl = MemHandler()
        x.logger.addHandler(hdl)
        x.logger.setLevel(logging.DEBUG)

        x.prop = 1
        y = x.prop
        with self.assertRaises(Exception):
            x.prop = None

        with self.assertRaises(Exception):
            x.properr

        self.assertEqual(hdl.history, ['Setting prop to 1',
                                       'prop was set to 1',
                                       'Getting prop',
                                       'Got 3 for prop',
                                       'Setting prop to None',
                                       'While setting prop to None: Arrrg!',
                                       'Getting properr',
                                       'While getting properr: GetArrrg!'])

    def test_log_config_fun(self):

        Dummy = define(lambda: props.LogProperty(log_values=lambda x: 2 * x if isinstance(x, int) else x), mixins.LogMixin)
        x = Dummy()

        hdl = MemHandler()
        x.logger.addHandler(hdl)
        x.logger.setLevel(logging.DEBUG)

        x.prop = 1
        y = x.prop
        with self.assertRaises(Exception):
            x.prop = None

        with self.assertRaises(Exception):
            x.properr

        self.assertEqual(hdl.history, ['Setting prop to 2',
                                       'prop was set to 2',
                                       'Getting prop',
                                       'Got 6 for prop',
                                       'Setting prop to None',
                                       'While setting prop to None: Arrrg!',
                                       'Getting properr',
                                       'While getting properr: GetArrrg!'])


    def test_lock(self):

        with self.assertRaises(Exception):
            define(props.LockProperty)

        Dummy = define(props.LockProperty, mixins.LockMixin)
        x = Dummy()

        # TODO better test locks
        x.prop = 3
        x.prop

    def test_transform_get(self):

        with self.assertRaises(Exception):
            define(props.TransformProperty)

        class C(mixins.StorageMixin, mixins.BaseLogMixin):

            _value = 4

            @props.TransformProperty(post_get=lambda x: 2*x)
            def prop(self):
                return self._value

            @prop.setter
            def prop(self, value):
                self._value = value

        c = C()
        self.assertEqual(c.prop, 2*c._value)
        c.prop = 1
        self.assertEqual(c.prop, 2*c._value)

    def test_transform_set(self):

        class C(mixins.StorageMixin, mixins.BaseLogMixin):

            _value = 4

            @props.TransformProperty(pre_set=lambda x: 3*x)
            def prop(self):
                return self._value

            @prop.setter
            def prop(self, value):
                self._value = value

        c = C()
        self.assertEqual(c.prop, c._value)
        c.prop = 3
        self.assertEqual(c._value, 9)

    def test_cache(self):

        # Defaults to False
        Dummy = define(lambda: props.ReadOnceProperty(),
                       mixins.CacheMixin, mixins.BaseLogMixin, mixins.StorageMixin)
        x = Dummy()

        self.assertEqual(x.prop, 3)
        self.assertEqual(x.prop_gs, 8)
        self.assertEqual(x.recall('prop'), 3)
        self.assertEqual(x.recall(('prop', 'prop_gs')), dict(prop=3, prop_gs=8))

    def test_readonce(self):

        # Defaults to False
        Dummy = define(lambda: props.ReadOnceProperty(),
                       mixins.CacheMixin, mixins.BaseLogMixin, mixins.StorageMixin)
        x = Dummy()

        self.assertEqual(x.prop, 3)
        x._internal = 9
        self.assertEqual(x.prop, 9)


    def test_readonce_true(self):
        Dummy = define(lambda: props.ReadOnceProperty(read_once=True),
                       mixins.CacheMixin, mixins.BaseLogMixin, mixins.StorageMixin)
        x = Dummy()

        self.assertEqual(x.prop, 3)
        x._internal = 9
        self.assertEqual(x.prop, 3)

    def test_prevent_unnecesary_set(self):

        # Defaults to False
        Dummy = define(props.PreventUnnecessarySetProperty,
                       mixins.CacheMixin, mixins.BaseLogMixin, mixins.StorageMixin)
        x = Dummy()

        self.assertEqual(x.prop_gs, 8)
        x.prop_gs = 9
        self.assertEqual(x.prop_gs, 9)

        # Because we prevent unnecessary set but we change it under the hood
        # the value is not updated
        x._prop_gs = 0
        x.prop_gs = 9
        self.assertEqual(x.prop_gs, 0)


class TestPropertyConfig(unittest.TestCase):

    def test_config(self):

        class MyProp(props.NamedProperty):

            cfg = common.Config()

        class Dummy:

            @MyProp(cfg=1)
            def prop(self):
                return None

        self.assertEqual(Dummy.prop._config, dict(cfg=1))
        self.assertEqual(Dummy.prop._kwargs, dict(cfg=1))

    def test_config_missing(self):

        class MyProp(props.NamedProperty):

            cfg = common.Config()

        with self.assertRaises(TypeError):
            class Dummy:

                @MyProp()
                def prop(self):
                    return None

    def test_config_wrong1(self):

        class MyProp(props.NamedProperty):
            pass

        with self.assertRaises(TypeError):
            class Dummy:

                @MyProp(cfg2=20)
                def prop(self):
                    return None

    def test_config_wrong2(self):

        class MyProp(props.NamedProperty):

            cfg = common.Config()

        with self.assertRaises(TypeError):
            class Dummy:

                @MyProp(cfg2=20)
                def prop(self):
                    return None

    def test_config_default(self):

        class MyProp(props.NamedProperty):

            cfg = common.Config(default=42)

        class Dummy:

            @MyProp()
            def prop(self):
                return None

        self.assertEqual(Dummy.prop._config, dict(cfg=42))
        self.assertEqual(dict(Dummy.prop.config_iter(None)), dict(cfg=42))
        self.assertEqual(Dummy.prop._kwargs, {})

    def test_config_default_changed(self):

        class MyProp(props.NamedProperty):

            cfg = common.Config(default=42)

        class Dummy:

            @MyProp(cfg=43)
            def prop(self):
                return None

        self.assertEqual(Dummy.prop._config, dict(cfg=43))
        self.assertEqual(Dummy.prop._kwargs, dict(cfg=43))

    def test_config_values(self):

        class MyProp(props.NamedProperty):

            cfg = common.Config(valid_values=(True, False))

        class Dummy:

            @MyProp(cfg=True)
            def prop(self):
                return None

        with self.assertRaises(ValueError):
            class Dummy:

                @MyProp(cfg=42)
                def prop(self):
                    return None

    def test_config_types(self):

        class MyProp(props.NamedProperty):

            cfg = common.Config(valid_types=(int, ))

        class Dummy:

            @MyProp(cfg=True)
            def prop(self):
                return None

        with self.assertRaises(TypeError):
            class Dummy:

                @MyProp(cfg=32.2)
                def prop(self):
                    return None

    def test_config_check_func1(self):

        class MyProp(props.NamedProperty):

            cfg = common.Config(check_func=lambda x: x == 32)

        class Dummy:

            @MyProp(cfg=32)
            def prop(self):
                return None

        with self.assertRaises(ValueError):
            class Dummy:

                @MyProp(cfg=50)
                def prop(self):
                    return None

    def test_config_check_func2(self):

        def _check(x):
            if x == 32:
                return True
            raise AttributeError

        class MyProp(props.NamedProperty):

            cfg = common.Config(check_func=_check)

        class Dummy:

            @MyProp(cfg=32)
            def prop(self):
                return None

        with self.assertRaises(ValueError):
            class Dummy:

                @MyProp(cfg=50)
                def prop(self):
                    return None


class TestDocs(unittest.TestCase):

    def assertDocEqual(self, doc1, doc2):
        if doc1 is None or doc2 is None:
            self.assertEqual(doc1, doc2)

        doc1 = '\n'.join(d.strip() for d in doc1.split('\n') if d.strip())
        doc2 = '\n'.join(d.strip() for d in doc2.split('\n') if d.strip())
        self.assertEqual(doc1, doc2)

    def test_empty(self):

        def correct():
            """
            Other parameters
            ----------------
            cfg
            """

        class MyProp(props.NamedProperty):

            cfg = common.Config()

        self.assertDocEqual(MyProp.__doc__, correct.__doc__)


    def test_non_empty(self):

        def correct():
            """test

            Other parameters
            ----------------
            cfg
            """

        class MyProp(props.NamedProperty):
            """test

            """

            cfg = common.Config()

        self.assertDocEqual(MyProp.__doc__, correct.__doc__)

    def test_default(self):

        def correct():
            """test

            Other parameters
            ----------------
            cfg : (default=1)
            """

        class MyProp(props.NamedProperty):
            """test

            """

            cfg = common.Config(default=1)

        self.assertDocEqual(MyProp.__doc__, correct.__doc__)

    def test_values(self):

        def correct():
            """test

            Other parameters
            ----------------
            cfg : True or False
            """

        class MyProp(props.NamedProperty):
            """test

            """

            cfg = common.Config(valid_values=(True, False))

        self.assertDocEqual(MyProp.__doc__, correct.__doc__)

    def test_types(self):

        def correct():
            """test

            Other parameters
            ----------------
            cfg : int or float
            """

        class MyProp(props.NamedProperty):
            """test

            """

            cfg = common.Config(valid_types=(int, float))

        self.assertDocEqual(MyProp.__doc__, correct.__doc__)

    def test_doc(self):

        def correct():
            """test

            Other parameters
            ----------------
            cfg
                testing 123
            """

        class MyProp(props.NamedProperty):
            """test

            """

            cfg = common.Config(doc='testing 123')

        self.assertDocEqual(MyProp.__doc__, correct.__doc__)

    def test_all(self):

        def correct():
            """test

            Other parameters
            ----------------
            cfg : True or False (default=True)
                testing 123
            """

        class MyProp(props.NamedProperty):
            """test

            """

            cfg = common.Config(valid_values=(True, False), doc='testing 123', default=True)

        self.assertDocEqual(MyProp.__doc__, correct.__doc__)


# -*- coding: utf-8 -*-
"""
    pimpmyclass
    ~~~~~~~~~~~

    Pimp your class with awesome features.

    The central purpose of the library is to extend python properties to allow:

        get/set logging.
        get/set timing, success and failure stats.
        async locking.
        get/set coercion and conversion.
        value cache
        prevent unnecessary set.
        read once properties
        and more ...

    But most importantly, it allows owner specific configurations. Properties are class attributes,
    and therefore it is difficult to have a property which is, for exampled cached, in an object
    but not cached in another instance of the same class.

    The library also provides DictProperties: that is properties that can be accessed by key;
    and also methods!

    Each capability is isolated in individual classes allowing you to pick only what you need.

    These library started by refactoring code from Lantz 0.3 into a reusable and composable module.

    :copyright: 2019 by pimpmyclass Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from . import props, dictprops, methods, mixins
from .common import Config, InstanceConfig

__all__ = ['props', 'dictprops', 'methods', 'mixins', 'Config', 'InstanceConfig']



import unittest
import logging

from pimpmyclass import mixins, methods, common


class MemHandler(logging.Handler):



    def __init__(self):
        super().__init__()
        self.setFormatter(logging.Formatter(style='{'))
        self.history = list()

    def emit(self, record):
        self.history.append(self.format(record))


def define(proptype, *bases):

    class Dummy(*bases):

        logger_name = 'testing123'

        value = 3

        @proptype()
        def method(self):
            return self.value

        @proptype()
        def method2(self, n):
            return self.value * n

        @proptype()
        def method3(self, n, t):
            return self.value * n

    return Dummy


class TestMethods(unittest.TestCase):

    def test_timing(self):

        with self.assertRaises(Exception):
            define(methods.StatsMethod)

        Dummy = define(methods.StatsMethod, mixins.StorageMixin)
        x = Dummy()
        y = Dummy()

        def g(i):
            return getattr(i.__class__, 'method')

        self.assertIsNot(g(x).stats, g(y).stats)

        s = g(x).stats(x, 'call')
        self.assertEqual(s.count, 0)

        self.assertEqual(x.method(), 3)

        s = g(x).stats(x, 'call')
        self.assertEqual(s.count, 1)

        s = g(y).stats(y, 'call')
        self.assertEqual(s.count, 0)

    def test_lock(self):

        with self.assertRaises(Exception):
            Dummy = define(methods.LockMethod)

        Dummy = define(methods.LockMethod, mixins.LockMixin)
        x = Dummy()
        self.assertEqual(x.method(), 3)

    def test_async(self):

        Dummy = define(methods.LockMethod, mixins.AsyncMixin)
        Dummy.attach_async(Dummy.method)
        x = Dummy()
        self.assertEqual(x.method(), 3)
        self.assertEqual(x.method_async().result(), 3)

    def test_transformations(self):

        with self.assertRaises(Exception):
            Dummy = define(methods.TransformMethod)

        Dummy = define(methods.TransformMethod, mixins.StorageMixin, mixins.BaseLogMixin)
        x = Dummy()
        self.assertEqual(x.method(), 3)

        self.assertEqual(x.method2(2), 6)

        Dummy.method2.params = {'n': lambda x: 2*x}
        self.assertEqual(x.method2(2), 12)

    def test_transformations_param_ret(self):

        class Dummy(mixins.StorageMixin, mixins.BaseLogMixin):

            @methods.TransformMethod()
            @methods.TransformMethod.param('x', lambda o: 2 * o)
            @methods.TransformMethod.ret(lambda o: o / 9)
            def method(self, x, y):
                return x + y * 3

        d = Dummy()
        self.assertEqual(d.method(2, 3), (2 * 2 + 3 * 3) / 9)

    def test_transformations_param_tuple(self):

        class Dummy(mixins.StorageMixin, mixins.BaseLogMixin):

            @methods.TransformMethod()
            @methods.TransformMethod.param(('x', 'y'), lambda o: 2 * o)
            @methods.TransformMethod.ret(lambda o: o / 9)
            def method(self, x, y):
                return x + y * 3

        d = Dummy()
        self.assertEqual(d.method(2, 3), (2 * 2 + 2 * 3 * 3) / 9)

    def test_transformations1(self):

        class Dummy(mixins.StorageMixin, mixins.BaseLogMixin):

            @methods.TransformMethod(params=lambda o: 3 * o)
            def method(self, x):
                return x * 3

        d = Dummy()
        self.assertEqual(d.method(2), 2 * 3 * 3)

    def test_transformations_deprecated(self):

        with self.assertRaises(Exception):

            class Dummy(mixins.StorageMixin, mixins.BaseLogMixin):

                @methods.TransformMethod(params=lambda o: 3 * o)
                def method(self, x, y):
                    return x + y * 3

    def test_transformations_wrong_name(self):

        with self.assertRaises(Exception):

            class Dummy(mixins.StorageMixin, mixins.BaseLogMixin):

                @methods.TransformMethod()
                @methods.TransformMethod.param('w', lambda o: 2 * o)
                def method(self, x, y):
                    return x + y * 3

    def test_log(self):

        with self.assertRaises(Exception):
            define(methods.LogMethod)

        Dummy = define(methods.LogMethod, mixins.LogMixin)
        x = Dummy()

        hdl = MemHandler()
        x.logger.addHandler(hdl)
        x.logger.setLevel(logging.DEBUG)

        x.method()
        x.method2(3)

        self.assertEqual(hdl.history, ['Calling method',
                                       'method returned 3',
                                       'Calling method2 with ((3,), {}))',
                                       'method2 returned 9'])

    def test_log_config_false(self):

        Dummy = define(lambda: methods.LogMethod(log_values=False), mixins.LogMixin)
        x = Dummy()

        hdl = MemHandler()
        x.logger.addHandler(hdl)
        x.logger.setLevel(logging.DEBUG)

        x.method()
        x.method2(3)

        self.assertEqual(hdl.history, ['Calling method',
                                       "method returned <class 'int'>",
                                       "Calling method2 with ((<class 'int'>,), {}))",
                                       "method2 returned <class 'int'>"])


    def test_log_config_fun(self):

        Dummy = define(lambda: methods.LogMethod(log_values=lambda x: '%s %s' % (x, type(x))), mixins.LogMixin)
        x = Dummy()

        hdl = MemHandler()
        x.logger.addHandler(hdl)
        x.logger.setLevel(logging.DEBUG)

        x.method()
        x.method2(3)

        self.assertEqual(hdl.history, ['Calling method',
                                       "method returned 3 <class 'int'>",
                                       """Calling method2 with (("3 <class 'int'>",), {}))""",
                                       "method2 returned 9 <class 'int'>"])


class TestMethodsConfig(unittest.TestCase):

    def test_config(self):

        class MyMethod(methods.NamedMethod):

            cfg = common.Config()

        class Dummy:

            @MyMethod(cfg=1)
            def prop(self):
                return None

        self.assertEqual(Dummy.prop._config, dict(cfg=1))
        self.assertEqual(Dummy.prop._kwargs, dict(cfg=1))

    def test_config_missing(self):

        class MyMethod(methods.NamedMethod):

            cfg = common.Config()

        with self.assertRaises(TypeError):
            class Dummy:

                @MyMethod()
                def prop(self):
                    return None

    def test_config_wrong1(self):

        class MyMethod(methods.NamedMethod):
            pass

        with self.assertRaises(TypeError):
            class Dummy:

                @MyMethod(cfg2=20)
                def prop(self):
                    return None

    def test_config_wrong2(self):

        class MyMethod(methods.NamedMethod):

            cfg = common.Config()

        with self.assertRaises(TypeError):
            class Dummy:

                @MyMethod(cfg2=20)
                def prop(self):
                    return None

    def test_config_default(self):

        class MyMethod(methods.NamedMethod):

            cfg = common.Config(default=42)

        class Dummy:

            @MyMethod()
            def prop(self):
                return None

        self.assertEqual(Dummy.prop._config, dict(cfg=42))
        self.assertEqual(dict(Dummy.prop.config_iter(None)), dict(cfg=42))
        self.assertEqual(Dummy.prop._kwargs, {})

    def test_config_default_changed(self):

        class MyMethod(methods.NamedMethod):

            cfg = common.Config(default=42)

        class Dummy:

            @MyMethod(cfg=43)
            def prop(self):
                return None

        self.assertEqual(Dummy.prop._config, dict(cfg=43))
        self.assertEqual(Dummy.prop._kwargs, dict(cfg=43))

    def test_config_values(self):

        class MyMethod(methods.NamedMethod):

            cfg = common.Config(valid_values=(True, False))

        class Dummy:

            @MyMethod(cfg=True)
            def prop(self):
                return None

        with self.assertRaises(ValueError):
            class Dummy:

                @MyMethod(cfg=42)
                def prop(self):
                    return None

    def test_config_types(self):

        class MyMethod(methods.NamedMethod):

            cfg = common.Config(valid_types=(int, ))

        class Dummy:

            @MyMethod(cfg=True)
            def prop(self):
                return None

        with self.assertRaises(TypeError):
            class Dummy:

                @MyMethod(cfg=32.2)
                def prop(self):
                    return None

    def test_config_check_func1(self):

        class MyMethod(methods.NamedMethod):

            cfg = common.Config(check_func=lambda x: x == 32)

        class Dummy:

            @MyMethod(cfg=32)
            def prop(self):
                return None

        with self.assertRaises(ValueError):
            class Dummy:

                @MyMethod(cfg=50)
                def prop(self):
                    return None

    def test_config_check_func2(self):

        def _check(x):
            if x == 32:
                return True
            raise AttributeError

        class MyMethod(methods.NamedMethod):

            cfg = common.Config(check_func=_check)

        class Dummy:

            @MyMethod(cfg=32)
            def prop(self):
                return None

        with self.assertRaises(ValueError):
            class Dummy:

                @MyMethod(cfg=50)
                def prop(self):
                    return None



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from setuptools import setup

def read(filename):
    return open(filename, encoding='utf-8').read()


long_description = '\n\n'.join([read('README.md'),
                                read('AUTHORS'),
                                read('CHANGES')])

setup(
    name='pimpmyclass',
    version='0.4.3',
    description='Pimp your Class/Property/Methods with useful functionality',
    long_description=long_description,
    keywords='classes properties descriptors decorators',
    author='Hernan E. Grecco',
    author_email='hernan.grecco@gmail.com',
    url='https://github.com/hgrecco/pimpmyclass',
    test_suite='pimpmyclass.testsuite.testsuite',
    zip_safe=True,
    packages=['pimpmyclass'],
    include_package_data=True,
    license='BSD',
    python_requires = '>=3.6',
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: BSD License',
        'Operating System :: MacOS :: MacOS X',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering',
        'Topic :: Software Development :: Libraries',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ])



"""
    pimpmyproperty.mixins
    ~~~~~~~~~~~~~~~~~~~~~

    Mixins classes providing extra functionality to classes with pimped properties and methods

    - StorageMixin: Provides an instance based storage.
    - BaseLogMixin: Provides generic log methods.
    - LogMixin: Provides log methods using python logging module.
    - LockMixin: Provides an instance re-entrant lock.
    - AsyncMixin: Provides async capabilities via a queue.
    - CacheMixin: Provides a cache that can be accessed and invalidated.
    - ObservableMixin: Provides signals.

    :copyright: 2019 by pimpmyclass Authors, see AUTHORS for more details.
    :license: BSD, see LICENSE for more details.
"""

from concurrent import futures
import logging
import threading

from . import helpers


class StorageMixin:
    """Mixin class to be inherited by a class that uses a StorageProperty.

    Provides an instance specific storage space divided into namespaces.
    """

    # Instance specific storage, initialized upon first use.
    __storage = None


    @property
    def storage(self):
        """Storage for the instance.

        Returns
        -------
        dict-like object
        """
        if self.__storage is None:
            self.__storage = dict()

        return self.__storage


class BaseLogMixin:
    """Base Mixin class to be inherited by a class requiring logging.

    Derived classes can specify the logger_name by overriding the logger name variable.

    Optionally, an instance of logging.Logger can be attached after instantiation
    using the logger attribute.

    Extra information to each log record can be added permanently using the
    logger extra attribute.
    """

    def log(self, level, msg, *args, **kwargs):
        """Log with the integer severity 'level'
        on the logger corresponding to this class.

        Must be implemented by classes actually logging something.

        Parameters
        ----------
        level :
            severity level for this event.
        msg :
            message to be logged (can contain PEP3101 formatting codes)
        *args :
            arguments passed to logger.log
        **kwargs :
            keyword arguments passed to the logger.log

        See Also
        --------
        log_info, log_debug, log_error, log_warning, log_critical
        """

    def log_info(self, msg, *args, **kwargs):
        """Log with the severity 'INFO' on the logger corresponding to this instance.

        See Also
        --------
        log, log_debug, log_error, log_warning, log_critical
        """
        self.log(logging.INFO, msg, *args, **kwargs)

    def log_debug(self, msg, *args, **kwargs):
        """Log with the severity 'DEBUG' on the logger corresponding to this instance.

        See Also
        --------
        log, log_info, log_error, log_warning, log_critical
        """

        self.log(logging.DEBUG, msg, *args, **kwargs)

    def log_error(self, msg, *args, **kwargs):
        """Log with the severity 'ERROR' on the logger corresponding to this instance.

        See Also
        --------
        log, log_info, log_debug, log_warning, log_critical
        """

        self.log(logging.ERROR, msg, *args, **kwargs)

    def log_warning(self, msg, *args, **kwargs):
        """Log with the severity 'WARNING' on the logger corresponding to this instance.

        See Also
        --------
        log, log_info, log_debug, log_error, log_critical
        """

        self.log(logging.WARNING, msg, *args, **kwargs)

    def log_critical(self, msg, *args, **kwargs):
        """Log with the severity 'CRITICAL' on the logger corresponding to this instance.

        See Also
        --------
        log, log_info, log_debug, log_error, log_warning
        """

        self.log(logging.CRITICAL, msg, *args, **kwargs)


class LogMixin(BaseLogMixin):
    """Mixin class to be inherited by a class requiring logging to Python std logging.

    Derived classes can specify the logger_name by overriding the logger name variable.

    Optionally, an instance of logging.Logger can be attached after instantiation
    using the logger attribute.

    Extra information to each log record can be added permanently using the
    logger extra attribute.
    """

    __logger = None
    __logger_extra = None

    _get_logger = logging.getLogger

    logger_name = None

    @property
    def logger(self):
        if self.__logger is None:
            self.__logger = self.__class__._get_logger(self.logger_name)

        return self.__logger

    @logger.setter
    def logger(self, logger_instance):
        self.__logger = logger_instance

    @property
    def logger_extra(self):
        return self.__logger_extra

    @logger_extra.setter
    def logger_extra(self, dictlike):
        self.__logger_extra = dictlike

    def log(self, level, msg, *args, **kwargs):
        """Log with the integer severity 'level'
        on the logger corresponding to this class.

        Parameters
        ----------
        level :
            severity level for this event.
        msg :
            message to be logged (can contain PEP3101 formatting codes)
        *args :
            arguments passed to logger.log
        **kwargs :
            keyword arguments passed to the logger.log

        See Also
        --------
        log_info, log_debug, log_error, log_warning, log_critical
        """
        if self.__logger_extra:
            self.logger.log(level, msg, *args,
                            extra=dict(self.__logger_extra, **kwargs))
        else:
            self.logger.log(level, msg, *args, **kwargs)


class LockMixin:
    """Mixin class to be inherited by a class requiring a reentrant lock.
    """

    __async_lock = None

    @property
    def lock(self):
        if self.__async_lock is None:
            self.__async_lock = threading.RLock()

        return self.__async_lock


class AsyncMixin(LockMixin):
    """Mixin class to be inherited by a class requiring async operations.

    Async operations are implemented using a single worker Thread Pool Executor
    that returns futures.

    The number of pending tasks can be found in `async_pending`.
    """

    __async_executor = None
    __async_pending = 0

    async_pending = property(lambda self: self.__async_pending)

    @property
    def async_executor(self):
        if self.__async_executor is None:
            self.__async_executor = futures.ThreadPoolExecutor(max_workers=1)

        return self.__async_executor

    def _async_done(self, fut):
        with self.lock:
            self.__async_pending -= 1
        return fut

    def _async_submit(self, fn, *args, **kwargs):
        with self.lock:
            self.__async_pending += 1
        fut = self.async_executor.submit(fn, *args, **kwargs)
        fut.add_done_callback(self._async_done)
        return fut

    def _async_submit_by_name(self, fname, *args, **kwargs):
        return self._async_submit(getattr(self, fname), *args, **kwargs)

    @classmethod
    def attach_async(cls, func):
        def async_func(self, *args, **kwargs):
            return self._async_submit_by_name(func.name, *args, **kwargs)

        async_func = async_func
        async_func.__doc__ = helpers.prepend_to_docstring('(Async) ', func.__doc__)
        setattr(cls, func.name + '_async', async_func)
        return async_func


class CacheMixin:

    _cache_unset_value = None

    def recall(self, keys):
        """Return the last value seen for a CacheProperty or a collection of CacheProperties.

        Parameters
        ----------
        keys : str or iterable of str, optional
            Name of the CacheProperty or properties to recall.

        Returns
        -------
        value of the CacheProperty or dict mapping CacheProperty to values.

        """

        if isinstance(keys, (list, tuple, set)):
            return {key: getattr(self.__class__, key).recall(self) for key in keys}
        return getattr(self.__class__, keys).recall(self)


class ObservableMixin:

    _observer_signal_init = None


